@@ -20,10 +20,9 @@ from pyrsistent import (
     PClass,
     optional,
     pmap_field,
-    pvector_field,
     pvector, )
 from boltons.funcutils import wraps
-from six import text_type as unicode, integer_types, PY3
+from six import text_type as unicode, PY3
 
 from ._message import (
     Message,
@@ -50,15 +49,24 @@ class _ExecutionContext(threading.local):
 
     The context is thread-specific, but can be made e.g. coroutine-specific by
     overriding C{get_sub_context}.
+
+    If we ever change this again we might want to make a ``threading.local``
+    subclass a sub-object?  Or come up with better API for thread-locals.
     """
 
     def __init__(self):
+        """
+        Every time an attribute is looked up from a new thread, this will be
+        called again for that thread, because this is a ``threading.local``
+        subclass.
+        """
         self._main_stack = []
         self.get_sub_context = lambda: None
 
     def _get_stack(self):
         """
-        Get the stack for the current asyncio Task.
+        Get the stack for the current thread, or the sub-stack if sub-stacks
+        are enabled.
         """
         stack = self.get_sub_context()
         if stack is None:
@@ -92,23 +100,54 @@ class _ExecutionContext(threading.local):
         return stack[-1]
 
 
-_context = _ExecutionContext()
-current_action = _context.current
+class _ECOwner(object):
+    """Owner of the global execution context singleton.
 
+    It allows setting-once-only a replacement class for the default
+    L{_ExecutionContext}, so different sub-contexts (e.g. asyncio and
+    generators) don't stomp on each other.
 
-def use_asyncio_context():
+    @ivar context: The current global L{_ExecutionContext}.  Don't set it
+        directly, only get it!  You can use C{set} to set it.
     """
-    Use a logging context that is tied to the current asyncio coroutine.
 
-    Call this first thing, before doing any other logging.
+    def __init__(self):
+        self.reset()
 
-    Does not currently support event loops other than asyncio.
+    def reset(self):
+        """Reset to default context, to be used by tests only."""
+        self.context = _ExecutionContext()
+        self._set = False
+
+    def set(self, context_class):
+        """Set a new context of the given class.
+
+        If the same class as current one, no changes are made.
+
+        @raises C{RuntimeError}: If the context has already been set to a
+            different class.
+        """
+        if self.context.__class__ == context_class:
+            return
+        if self._set:
+            raise RuntimeError(
+                "Context class already set to " + str(self.context.__class__)
+            )
+        self.context = context_class()
+        self._set = True
+
+
+_context_owner = _ECOwner()
+
+
+def current_action():
     """
-    from ._asyncio import AsyncioContext
-    _context.get_sub_context = AsyncioContext().get_stack
+    @return: The current C{Action} in context, or C{None} if none were set.
+    """
+    return _context_owner.context.current()
 
 
-class TaskLevel(PClass):
+class TaskLevel(object):
     """
     The location of a message within the tree of actions of a task.
 
@@ -117,23 +156,45 @@ class TaskLevel(PClass):
         3]} indicates this is the third message within an action which is
         the second item in the task.
     """
+    def __init__(self, level):
+        self._level = level
 
-    level = pvector_field(integer_types)
+    def as_list(self):
+        """Return the current level.
 
-    # PClass really ought to provide this ordering facility for us:
-    # tobgu/pyrsistent#45.
+        @return: List of integers.
+        """
+        return self._level[:]
+
+    # Backwards compatibility:
+    @property
+    def level(self):
+        return pvector(self._level)
 
     def __lt__(self, other):
-        return self.level < other.level
+        return self._level < other._level
 
     def __le__(self, other):
-        return self.level <= other.level
+        return self._level <= other._level
 
     def __gt__(self, other):
-        return self.level > other.level
+        return self._level > other._level
 
     def __ge__(self, other):
-        return self.level >= other.level
+        return self._level >= other._level
+
+    def __eq__(self, other):
+        if other.__class__ != TaskLevel:
+            return False
+        return self._level == other._level
+
+    def __ne__(self, other):
+        if other.__class__ != TaskLevel:
+            return True
+        return self._level != other._level
+
+    def __hash__(self):
+        return hash(tuple(self._level))
 
     @classmethod
     def fromString(cls, string):
@@ -152,7 +213,7 @@ class TaskLevel(PClass):
 
         @return: L{unicode} representation of the L{TaskLevel}.
         """
-        return "/" + "/".join(map(unicode, self.level))
+        return "/" + "/".join(map(unicode, self._level))
 
     def next_sibling(self):
         """
@@ -161,7 +222,9 @@ class TaskLevel(PClass):
 
         @return: L{TaskLevel} which follows this one.
         """
-        return TaskLevel(level=self.level.set(-1, self.level[-1] + 1))
+        new_level = self._level[:]
+        new_level[-1] += 1
+        return TaskLevel(level=new_level)
 
     def child(self):
         """
@@ -169,7 +232,9 @@ class TaskLevel(PClass):
 
         @return: L{TaskLevel} which is the first child of this one.
         """
-        return TaskLevel(level=self.level.append(1))
+        new_level = self._level[:]
+        new_level.append(1)
+        return TaskLevel(level=new_level)
 
     def parent(self):
         """
@@ -178,9 +243,9 @@ class TaskLevel(PClass):
 
         @return: L{TaskLevel} which is the parent of this one.
         """
-        if not self.level:
+        if not self._level:
             return None
-        return TaskLevel(level=self.level[:-1])
+        return TaskLevel(level=self._level[:-1])
 
     def is_sibling_of(self, task_level):
         """
@@ -397,11 +462,11 @@ class Action(object):
         """
         Run the given function with this L{Action} as its execution context.
         """
-        _context.push(self)
+        _context_owner.context.push(self)
         try:
             return f(*args, **kwargs)
         finally:
-            _context.pop()
+            _context_owner.context.pop()
 
     def addSuccessFields(self, **fields):
         """
@@ -422,25 +487,25 @@ class Action(object):
 
         The action does NOT finish when the context is exited.
         """
-        _context.push(self)
+        _context_owner.context.push(self)
         try:
             yield self
         finally:
-            _context.pop()
+            _context_owner.context.pop()
 
     # Python context manager implementation:
     def __enter__(self):
         """
         Push this action onto the execution context.
         """
-        _context.push(self)
+        _context_owner.context.push(self)
         return self
 
     def __exit__(self, type, exception, traceback):
         """
         Pop this action off the execution context, log finish message.
         """
-        _context.pop()
+        _context_owner.context.pop()
         self.finish(exception)
 
 
