@@ -31,7 +31,7 @@ from .._action import (
     ACTION_STATUS_FIELD, ACTION_TYPE_FIELD, FAILED_STATUS, STARTED_STATUS,
     SUCCEEDED_STATUS, DuplicateChild, InvalidStartMessage, InvalidStatus,
     TaskLevel, WrittenAction, WrongActionType, WrongTask, WrongTaskLevel,
-    TooManyCalls, log_call
+    TooManyCalls, log_call, _context_owner
 )
 from .._message import (
     EXCEPTION_FIELD,
@@ -46,7 +46,6 @@ from .._validation import ActionType, Field, _ActionSerializers
 from ..testing import assertContainsFields, capture_logging
 from ..parse import Parser
 from .. import (
-    _action,
     add_destination,
     remove_destination,
     register_exception_extractor,
@@ -66,7 +65,6 @@ from .strategies import (
     union,
     written_from_pmap,
 )
-import eliot
 
 
 class ExecutionContextTests(TestCase):
@@ -178,14 +176,76 @@ class ExecutionContextTests(TestCase):
         self.assertEqual(valuesInThread, [second])
         self.assertIs(ctx.current(), first)
 
-    def test_globalInstance(self):
+
+class ContextOwnerTests(TestCase):
+    """Tests for the shared global context."""
+
+    def test_reset(self):
         """
-        A global L{_ExecutionContext} is exposed in the L{eliot._action}
-        module.
+        Resetting resets to a normal C{_ExecutionContext}.
         """
-        self.assertIsInstance(_action._context, _ExecutionContext)
-        self.assertEqual(_action.current_action, _action._context.current)
-        self.assertEqual(eliot.current_action, _action._context.current)
+        _context_owner.context = object()
+        _context_owner.reset()
+        self.assertIsInstance(_context_owner.context, _ExecutionContext)
+
+    def test_current_action(self):
+        """
+        L{current_action} returns the current value of the current global
+        context.
+        """
+        self.addCleanup(_context_owner.reset)
+        _context_owner.context.push("A")
+        self.assertEqual(current_action(), "A")
+        _context_owner.context.push("B")
+        self.assertEqual(current_action(), "B")
+        _context_owner.reset()
+        self.assertEqual(current_action(), None)
+        _context_owner.context.push("C")
+        self.assertEqual(current_action(), "C")
+
+    def test_set(self):
+        """
+        It's possible to set a new C{_ExecutionContext} once.
+        """
+        self.addCleanup(_context_owner.reset)
+
+        class A(_ExecutionContext):
+            pass
+
+        _context_owner.set(A)
+        self.assertIsInstance(_context_owner.context, A)
+
+    def test_set_twice(self):
+        """
+        Setting the same new C{_ExecutionContext} twice is no-op.
+        """
+        self.addCleanup(_context_owner.reset)
+
+        class A(_ExecutionContext):
+            pass
+
+        _context_owner.set(A)
+        a = _context_owner.context
+        _context_owner.set(A)
+        # Same object as before:
+        self.assertIs(_context_owner.context, a)
+
+    def test_set_twice_different_class(self):
+        """
+        Setting twice with different C{_ExecutionContext} classes raises a
+        C{RuntimeError}.
+        """
+        self.addCleanup(_context_owner.reset)
+
+        class A(_ExecutionContext):
+            pass
+
+        class B(_ExecutionContext):
+            pass
+
+        _context_owner.set(A)
+        with self.assertRaises(RuntimeError):
+            _context_owner.set(B)
 
 
 class ActionTests(TestCase):
@@ -960,6 +1020,37 @@ class TaskLevelTests(TestCase):
             )
         )
 
+    def test_equality(self):
+        """
+        L{TaskChild} correctly implements equality and hashing.
+        """
+        a = TaskLevel(level=[1, 2])
+        a2 = TaskLevel(level=[1, 2])
+        b = TaskLevel(level=[2, 999])
+        self.assertTrue(
+            all(
+                [
+                    a == a2,
+                    a2 == a,
+                    a != b,
+                    b != a,
+                    not b == a,
+                    not a == b,
+                    not a == 1,
+                    a != 1,
+                    hash(a) == hash(a2),
+                    hash(b) != hash(a)
+                ]
+            )
+        )
+
+    def test_as_list(self):
+        """
+        L{TaskChild.as_list} returns the level.
+        """
+        self.assertEqual(TaskLevel(level=[1, 2, 3]).as_list(),
+                         [1, 2, 3])
+
     @given(lists(task_level_indexes))
     def test_parent_of_child(self, base_task_level):
         """
