@@ -8,7 +8,7 @@ import sys
 from functools import wraps
 
 try:
-    from twisted.internet.defer import Deferred, succeed, fail
+    from twisted.internet.defer import Deferred, succeed, fail, returnValue
     from twisted.trial.unittest import TestCase
     from twisted.python.failure import Failure
     from twisted.logger import globalLogPublisher
@@ -20,13 +20,18 @@ else:
     # logwriter.py causes a failure:
     from ..twisted import (
         DeferredContext, AlreadyFinished, _passthrough, redirectLogsForTrial,
-        _RedirectLogsForTrial, TwistedDestination
+        _RedirectLogsForTrial, TwistedDestination,
+        inline_callbacks,
     )
 
-from .._action import start_action, current_action, Action, TaskLevel
+from .test_generators import assert_expected_action_tree
+
+from .._action import (
+    start_action, current_action, Action, TaskLevel, _context_owner
+)
 from .._output import MemoryLogger, Logger
 from .._message import Message
-from ..testing import assertContainsFields
+from ..testing import assertContainsFields, capture_logging
 from .. import removeDestination, addDestination
 from .._traceback import write_traceback
 from .common import FakeSys
@@ -121,6 +126,45 @@ class DeferredContextTests(TestCase):
         result.errback(RuntimeError())
         self.assertEqual(called, [(1, 2)])
 
+    @withActionContext
+    def test_addCallbacksWithOnlyCallback(self):
+        """
+        L{DeferredContext.addCallbacks} can be called with a single argument, a
+        callback function, and passes it to the wrapped L{Deferred}'s
+        C{addCallbacks}.
+        """
+        called = []
+        def f(value):
+            called.append(value)
+
+        result = Deferred()
+        context = DeferredContext(result)
+        context.addCallbacks(f)
+        result.callback(0)
+        self.assertEqual(called, [0])
+
+    @withActionContext
+    def test_addCallbacksWithOnlyCallbackErrorCase(self):
+        """
+        L{DeferredContext.addCallbacks} can be called with a single argument, a
+        callback function, and passes a pass-through errback to the wrapped
+        L{Deferred}'s C{addCallbacks}.
+        """
+        called = []
+        def f(value):
+            called.append(value)
+
+        class ExpectedException(Exception):
+            pass
+
+        result = Deferred()
+        context = DeferredContext(result)
+        context.addCallbacks(f)
+        result.errback(Failure(ExpectedException()))
+        self.assertEqual(called, [])
+        # The assertion is inside `failureResultOf`.
+        self.failureResultOf(result, ExpectedException)
+
     @withActionContext
     def test_addCallbacksReturnSelf(self):
         """
@@ -634,3 +678,207 @@ class TwistedDestinationTests(TestCase):
         self.assertEqual(
             writtenToTwisted, [("critical", written[0])]
         )
+
+
+class InlineCallbacksTests(TestCase):
+    """Tests for C{inline_callbacks}."""
+
+    # Get our custom assertion failure messages *and* the standard ones.
+    longMessage = True
+
+    def setUp(self):
+        # The generator sub-contxt is enabled automatically, so reset back to
+        # normal:
+        self.addCleanup(_context_owner.reset)
+
+    def _a_b_test(self, logger, g):
+        """A yield was done in between messages a and b inside C{inline_callbacks}."""
+        with start_action(action_type=u"the-action"):
+            self.assertIs(
+                None,
+                self.successResultOf(g()),
+            )
+        assert_expected_action_tree(
+            self,
+            logger,
+            u"the-action", [
+                u"a",
+                u"yielded",
+                u"b",
+            ],
+        )
+
+    @capture_logging(None)
+    def test_yield_none(self, logger):
+        def g():
+            Message.log(message_type=u"a")
+            yield
+            Message.log(message_type=u"b")
+        g = inline_callbacks(g, debug=True)
+
+        self._a_b_test(logger, g)
+
+    @capture_logging(None)
+    def test_yield_fired_deferred(self, logger):
+        def g():
+            Message.log(message_type=u"a")
+            yield succeed(None)
+            Message.log(message_type=u"b")
+        g = inline_callbacks(g, debug=True)
+
+        self._a_b_test(logger, g)
+
+    @capture_logging(None)
+    def test_yield_unfired_deferred(self, logger):
+        waiting = Deferred()
+
+        def g():
+            Message.log(message_type=u"a")
+            yield waiting
+            Message.log(message_type=u"b")
+        g = inline_callbacks(g, debug=True)
+
+        with start_action(action_type=u"the-action"):
+            d = g()
+            self.assertNoResult(waiting)
+            waiting.callback(None)
+            self.assertIs(
+                None,
+                self.successResultOf(d),
+            )
+        assert_expected_action_tree(
+            self,
+            logger,
+            u"the-action", [
+                u"a",
+                u"yielded",
+                u"b",
+            ],
+        )
+
+    @capture_logging(None)
+    def test_returnValue(self, logger):
+        result = object()
+
+        @inline_callbacks
+        def g():
+            if False:
+                yield
+            returnValue(result)
+
+        with start_action(action_type=u"the-action"):
+            d = g()
+            self.assertIs(
+                result,
+                self.successResultOf(d),
+            )
+
+        assert_expected_action_tree(
+            self,
+            logger,
+            u"the-action",
+            [],
+        )
+
+    @capture_logging(None)
+    def test_returnValue_in_action(self, logger):
+        result = object()
+
+        @inline_callbacks
+        def g():
+            if False:
+                yield
+            with start_action(action_type=u"g"):
+                returnValue(result)
+
+        with start_action(action_type=u"the-action"):
+            d = g()
+            self.assertIs(
+                result,
+                self.successResultOf(d),
+            )
+
+        assert_expected_action_tree(
+            self,
+            logger,
+            u"the-action", [
+                {u"g": []},
+            ],
+        )
+
+    @capture_logging(None)
+    def test_nested_returnValue(self, logger):
+        result = object()
+        another = object()
+
+        def g():
+            d = h()
+            # Run h through to the end but ignore its result.
+            yield d
+            # Give back _our_ result.
+            returnValue(result)
+        g = inline_callbacks(g, debug=True)
+
+        def h():
+            yield
+            returnValue(another)
+        h = inline_callbacks(h, debug=True)
+
+        with start_action(action_type=u"the-action"):
+            d = g()
+            self.assertIs(
+                result,
+                self.successResultOf(d),
+            )
+
+        assert_expected_action_tree(
+            self,
+            logger,
+            u"the-action", [
+                u"yielded",
+                u"yielded",
+            ],
+        )
+
+    @capture_logging(None)
+    def test_async_returnValue(self, logger):
+        result = object()
+        waiting = Deferred()
+
+        @inline_callbacks
+        def g():
+            yield waiting
+            returnValue(result)
+
+        with start_action(action_type=u"the-action"):
+            d = g()
+            waiting.callback(None)
+            self.assertIs(
+                result,
+                self.successResultOf(d),
+            )
+
+    @capture_logging(None)
+    def test_nested_async_returnValue(self, logger):
+        result = object()
+        another = object()
+
+        waiting = Deferred()
+
+        @inline_callbacks
+        def g():
+            yield h()
+            returnValue(result)
+
+        @inline_callbacks
+        def h():
+            yield waiting
+            returnValue(another)
+
+        with start_action(action_type=u"the-action"):
+            d = g()
+            waiting.callback(None)
+            self.assertIs(
+                result,
+                self.successResultOf(d),
+            )
